#!/usr/bin/env python
# encoding: utf8
"""Insert Tower data in to an InfluxDB for real time use.

"""
import logging
import logging.handlers
import time
import sys
import requests
from metobscommon import influxdb
from aosstower.level_00.parser import read_frames
from metobscommon.util import calc
from metobscommon.util.nc import calculate_wind_gust
import numpy as np
import pandas as pd
import warnings
from datetime import timedelta

LOG = logging.getLogger(__name__)
# map station name to InfluxDB tags
STATIONS = {
    "AOSS Tower": {"inst": "tower", "site": "aoss"},
}
# parser symbols to influxdb symbols
SYMBOL_CONVERSIONS = {
    'stamp': 'timestamp',
    'box_temp': 'box_temp',
    'box_pressure': 'box_pressure',
    'paro_air_temp_period': 'paro_air_temp_period',
    'paro_pressure_period': 'paro_pressure_period',
    'paro_air_temp': 'paro_air_temp',
    'pressure': 'pressure',
    'paro_cal_sig': 'paro_cal_sig',
    'box_rh': 'box_rh',
    'box_air_temp': 'box_air_temp',
    'air_temp_2': 'air_temp_2',
    'air_temp_3': 'air_temp_3',
    'air_temp_4': 'air_temp_4',
    'air_temp_5': 'air_temp_5',
    'wind_speed': 'wind_speed',
    'wind_dir': 'wind_dir',
    'rh_shield_freq': 'rh_shield_freq',
    'rh': 'rel_hum',
    'air_temp_6_3m': 'air_temp_6_3m',
    'dewpoint': 'dewpoint',
    'rtd_shield_freq': 'rtd_shied_freq',
    'air_temp': 'air_temp',
    'solar_flux': 'solar_flux',
    'precip': 'precip',
    'accum_precip': 'accum_precip',
    'altimeter': 'altimeter',
}
SYMBOLS = list(SYMBOL_CONVERSIONS.values())


class Updater(object):
    """Append weather record (taken as a dict) and do averages when enough data is ready.

    At least 12 minutes of data is required to do averaging for gust_10m: 10 minutes of wind gusts,
    and wind gusts need 2 minutes of data to calculate.

    This class is created once at startup and calls rolling_average every time new data is available which tries to do
    averaging every 5 minutes of data added.
    """
    def __init__(self, data_interval=timedelta(seconds=5), submit_interval=timedelta(minutes=5)):
        """intervals are timedelta objects."""
        self.data = {}
        self.start_time = None
        self.data_interval = data_interval.total_seconds()
        self.submit_interval = submit_interval.total_seconds()

    def rolling_average(self, record):
        # Appending to a DataFrame is slow. Instead, this adds to a dict in chunks and passes it to the DataFrame.
        time_mask = True
        if self.data.get('timestamp') is not None:
            time_mask = self.data['timestamp'] > record['timestamp'] - timedelta(minutes=12)
        for key in record:
            if self.data.get(key) is None:
                self.data[key] = np.array([])
            self.data[key] = np.append(self.data[key][time_mask], record[key])

        reference = pd.datetime(1, 1, 1)
        if (record['timestamp'] - reference).total_seconds() % self.submit_interval == 0:
            submit = True
        elif (record['timestamp'] - reference).total_seconds() % self.submit_interval + self.data_interval > \
                self.submit_interval:
            submit = True
        else:
            submit = False
        # If data hits a submit_interval interval, return data.
        if submit:
            return self._calculate_averages()

    def _calculate_averages(self):
        KNOTS_9 = calc.knots_to_mps(9.)
        KNOTS_5 = calc.knots_to_mps(5.)
        frame = pd.DataFrame(self.data)
        frame.set_index('timestamp', inplace=True)
        frame.mask(frame == -99999., inplace=True)
        full_frame = frame.asfreq('{0}S'.format(self.data_interval))
        # Add wind direction components so we can average wind direction properly.
        frame['wind_east'], frame['wind_north'], _ = calc.wind_vector_components(frame['wind_speed'],
                                                                                 frame['wind_dir'])
        frame['wind_dir'] = calc.wind_vector_degrees(frame['wind_east'], frame['wind_north'])

        if 'air_temp' in frame and 'rh' in frame and ('dewpoint' in frame or 'dewpoint_mean' in frame):
            LOG.info("'dewpoint' is missing from the input file, will calculate "
                     "it from air temp and relative humidity")
            frame['dewpoint'] = calc.dewpoint(frame['air_temp'], frame['rh'])

        # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        # 2 minute rolling average.
        winds_frame_2m = frame[['wind_speed', 'wind_east', 'wind_north']].rolling('2T', closed='right').mean()
        frame['wind_speed_2m'] = winds_frame_2m['wind_speed']
        frame['wind_dir_2m'] = calc.wind_vector_degrees(winds_frame_2m['wind_east'], winds_frame_2m['wind_north'])
        frame_2 = frame[frame.index > frame.index[-1] - timedelta(minutes=2)]
        full_2 = full_frame[full_frame.index > full_frame.index[-1] - timedelta(minutes=2)]
        # Makes 2 minute averages nans if given less than 2 minutes of data.
        if len(full_2) < 120 / self.data_interval or len(frame_2) != len(full_2):
            frame['wind_speed_2m'].mask(frame['wind_speed_2m'] > -1., inplace=True)
            frame['wind_dir_2m'].mask(frame['wind_dir_2m'] > -1., inplace=True)
        # 1 minute rolling peaks
        wind_peak_1m = frame['wind_speed'].rolling(window='1T', closed='right').max()
        # criteria for a fast wind to be considered a wind gust
        gust_mask = (winds_frame_2m['wind_speed'] >= KNOTS_9) &\
                    (wind_peak_1m >= winds_frame_2m['wind_speed'] + KNOTS_5)
        frame['gust_1m'] = wind_peak_1m.mask(~gust_mask)
        frame['gust_10m'] = calculate_wind_gust(frame['wind_speed'], winds_frame_2m['wind_speed'])
        # Makes 10 minute gusts before 12 minutes nans because data is insufficient.
        if len(full_frame) < 720 / self.data_interval or len(frame) != len(full_frame):
            frame['gust_10m'].mask(frame['gust_10m'] > -1., inplace=True)
        return frame.fillna(value=np.nan)


def convert_to_influx_frame(record_gen, symbols, debug=False):
    for idx, record in enumerate(record_gen):
        LOG.info("Inserting records for frame %d", idx)
        record = {symbols[k] or k: v for k, v in record.items() if k in symbols}
        if debug:
            print(idx, record)
            continue
        yield record


def construct_url(data):
    for key, val in data.items():
        if val is None or isinstance(val, float) and np.isnan(val):
            data[key] = ''
    return ('http://weatherstation.wunderground.com/weatherstation/updateweatherstation.php?'
            'ID={ID}&'
            'PASSWORD={PASSWORD}&'
            'action=updateraw&'
            'dateutc={dateutc}&'
            'winddir={winddir}&'
            'winddir_avg2m={winddir_avg2m}&'
            'windspeedmph={windspeedmph}&'
            'windspdmph_avg2m={windspdmph_avg2m}&'
            'windgustmph={windgustmph}&'
            'windgustmph_10m={windgustmph_10m}&'
            'humidity={humidity}&'
            'tempf={tempf}&'
            'baromin={baromin}&'
            'dewptf={dewptf}&'
            'solarradiation={solarradiation}&'
            'rainin={rainin}&'
            'dailyrainin={dailyrainin}&'
            'softwaretype=SSEC-RIG').format(**data)


def get_url_data(avg, wu_id, wu_pw):
    # Information on what paramaters that can be sent:
    # https://feedback.weather.com/customer/en/portal/articles/2924682-pws-upload-protocol?b_id=17298
    # For timestamp, want YYYY-MM-DD+hh:mm:ss of last dataset that was averaged, rounded up to nearest minute.
    timestamp = avg.index[-1].isoformat('+')
    wind_dir = avg['wind_dir'][-1]
    wind_dir_2m = avg['wind_dir_2m'][-1]
    rel_hum = avg['rel_hum'][-1]
    solar_flux = avg['solar_flux'][-1]
    precip = avg['precip'][-1]
    accum_precip = avg['accum_precip'][-1]
    # Converts from m/s to mph.
    wind_speed = avg['wind_speed'][-1] * 2.23694
    wind_speed_2m = avg['wind_speed_2m'][-1] * 2.23694
    gust_1m = avg['gust_1m'][-1] * 2.23694
    gust_10m = avg['gust_10m'][-1] * 2.23694
    # Converts degrees Celsius to degrees Fahrenheit
    air_temp = avg['air_temp'][-1] * 9. / 5. + 32.
    dewpoint = avg['dewpoint'][-1] * 9. / 5. + 32.
    # hpa to barometric pressure inches
    pressure = avg['pressure'][-1] * 0.02952998016471232
    return {'ID': wu_id, 'PASSWORD': wu_pw, 'dateutc': timestamp, 'winddir': wind_dir, 'winddir_avg2m': wind_dir_2m,
            'windspeedmph': wind_speed, 'windspdmph_avg2m': wind_speed_2m, 'windgustmph': gust_1m,
            'windgustmph_10m': gust_10m, 'humidity': rel_hum, 'tempf': air_temp, 'baromin': pressure,
            'dewptf': dewpoint, 'solarradiation': solar_flux, 'rainin': precip, 'dailyrainin': accum_precip}


def main():
    import argparse
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--logfn', help='Log to rotating file (Not Implemented)')
    parser.add_argument('--debug', action='store_true',
                        help='Don\'t submit records to the database, print them to stdout')
    parser.add_argument('-t', '--tail', action='store_true',
                        help=('Tail file forever, not returning. This will start at the end '
                              'of the file and insert any new data added after starting'))
    parser.add_argument('--ldmp', action='store_true',
                        help='Treat `src` file as a station name and read records from LoggerNet LDMP server (port: 1024)')
    parser.add_argument('--tables', nargs='*', default=['1'],
                        help="LoggerNet LDMP tables to read in")
    parser.add_argument("--host", default=influxdb.DB_HOST,
                        help="Hostname of database connection")
    parser.add_argument("--port", default=influxdb.DB_PORT,
                        help="Port of database connection")
    parser.add_argument("--dbname", default=influxdb.DB_NAME,
                        help="Name of database to modify")
    parser.add_argument('-s', '--station', dest='station', default='AOSS Tower', choices=STATIONS.keys(),
                        help='Name of station to use to determine symbols')
    parser.add_argument('-v', '--verbose', dest='verbosity', action="count", default=0,
                        help='each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG')
    parser.add_argument('--sleep-interval', type=float,
                        help="Seconds to wait between submitting each record")
    parser.add_argument('--weather-underground', action='store_true',
                        help="Send new records to wunderground.com")
    parser.add_argument('--wu-id', default='KWIMADIS52',
                        help='Weather underground station ID')
    parser.add_argument('--wu-password-file', default='/home/metobs/wunderground_password.txt',
                        help='File containing the password for the weather underground upload')
    parser.add_argument('--bulk', type=int, default=1,
                        help="Number of records to buffer before sending to "
                             "the database. For large inserts this should be "
                             "between 5000 to 10000. Default: 1")
    parser.add_argument('src', help='Level 0 raw data file or station name '
                                    'for LDMP reading')

    args = parser.parse_args()
    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=levels[min(3, args.verbosity)])

    LOG.info("source: %s", args.src)
    LOG.info("tail: %s", args.tail)
    station_tags = STATIONS[args.station]
    symbols = SYMBOL_CONVERSIONS

    wu_pw = None
    if args.weather_underground:
        wu_pw = open(args.wu_password_file, 'r').read().strip()

    if args.ldmp:
        from aosstower.level_00.parser import LDMPGenerator
        record_gen = LDMPGenerator(args.src, args.tables)
    else:
        src = open(args.src, "r")
        record_gen = read_frames(src, tail=args.tail)

    try:
        influx_gen = convert_to_influx_frame(record_gen, symbols, args.debug)
        influx_gen = influxdb.grouper(influx_gen, args.bulk)
        updater = Updater(submit_interval=timedelta(minutes=30))
        import time
        t0 = time.time()
        for record in influx_gen:
            # lines = influxdb.frame_records(record, **station_tags)
            # influxdb.insert(lines, host=args.host, port=args.port, dbname=args.dbname)
            # Record is in a list of size 1, but want just the record.
            avg = updater.rolling_average(record[0])
            # Once every 5 minutes: 0 through 295 seconds inclusive in 5 second intervals.
            if avg is not None:
                url = construct_url(get_url_data(avg, args.wu_id, wu_pw))
                print(url)
                if wu_pw and args.ldmp:
                    resp = requests.post(url)
                    if resp.status_code != 200:
                        warnings.warn('Data failed to upload to {0} with status code {1}: {2}'.format(
                            url, resp.status_code, resp.text))

            if args.sleep_interval:
                time.sleep(args.sleep_interval)
        t1 = time.time()
        print(t1 - t0)
    except (RuntimeError, ValueError, KeyError, requests.RequestException):
        raise
        if hasattr(record_gen, 'close'):
            record_gen.close()


if __name__ == "__main__":
    sys.exit(main())
